/************************************************************************************
MODULE     :   LocBusSlave        AUTHOR    :   Tullio Grassi

DESCRIPTION :
The Local Bus interfaces different components on a board. This bus includes the data
lines, the address lines, WRITE_b (active low) and several CSn_b (active low
ChipSelect: CS1_b, CS2_b, for all the  slave devices).
The protocol foresees that when the master drives CS_b low for a particular slave,
the slave must decode the address on the Loc_Address lines, and read or write the 
Loc_Data bus depending of the WRITE_b direction.
On a write cycle, the master set WRITE_b = 0, set the address and data on the Local Bus,
and the slave will latch the data when CS_b = 0. 
The master keeps the data valid for the whole duration of the CS_b pulse. 
On a read cycle, the master set WRITE_b = 1, set the address on the Local Bus and
latch the data on the rising edge of CS_b, as the slave must prepare the data 
during the CS_b pulse. The CS_b pulse width depends on the slower slave.
A value of 300 ns is usually fine. When a slave is not selected, its data port is on Hi-Z.
The design is fully synchronous with clk. 
LocalBus lines are syncronized (this adds a few ticks of delay for edge detection).
**********************************************************************************/

`include "define.v"
module LocBusSlave  (      
// these ports connect to the LocalBus
  input  LocCS_b,  LocWRITE_b,       
  input  [`LB_Addr_Size-1:0]  LocAddr,  // The Address bus.
  inout  [`REG_WIDTH-1:0]     LocData,  // The bidir Data bus 
// these ports connect to the rest of the FPGA
  input clk,  rst, DLL_lock, 
        run_status,  BunchEr, 
  input [2:0] UnlockCnt,
  input [7:0] LastTTCcommand,
  input [15:0] TTC_test0, TTC_test1, TTC_test2, TTC_test3,
               TTC_test4, TTC_test5, TTC_test6, TTC_test7,
               TTC_test8, TTC_test9, TTC_testA, TTC_testB,
               TTC_testC, TTC_testD, TTC_testE, TTC_testF,
   Pattern1Back, Pattern2Back, Pattern3Back, Pattern4Back,
   Pattern5Back, Pattern6Back, Pattern7Back, Pattern8Back,
   SnapFIFO1,  SnapFIFO2,  SnapFIFO3, SnapFIFO4,
   SnapFIFO5,  SnapFIFO6,  SnapFIFO7, SnapFIFO8,

  output  CounterModeSel, PatternMode, RepeatMode,
          ClockSelect, GateTTCcommands, StartStop_Spy, 
          VME_Hard_rst, VME_Soft_rst, 
          VME_Start, VME_Stop, VME_AsyncFifoRst,
  output reg [`BCN_WIDTH-1:0] MaxBCN ='hDEB,  
  output reg [`REG_WIDTH-1:0] SlbsID ='h5A, 
  output reg [8:1] PatternsWrEn, SnapRdEn,   
  output reg [23:0]  ChMask,        
   // outputs for debug 
  output reg Tri,  
  output reg [2:0]  LB_STATE,       
  output reg [`REG_WIDTH-1:0] Dout,    
  output     [`REG_WIDTH-1:0] Din    
);

  parameter 
  //  Parameters used for the STATEs of the LocBus FSM
  IDLE           =  0, // Waiting for the negative transition on LocCS_b
  RD_OR_WR       =  1, // check LocWRITE_b.  
  MASTER_WRITE   =  2, // If WR mode, then Store data
  MASTER_READ    =  3, // If RD mode, prepare data to be read
  END_OF_READ    =  4, // Latch data, release control lines, wait posedge of LocCS_b
  END_OF_WRITE   =  5, // Stop WR pulses and wait for posedge of LocCS_b
  END_CYCLE      =  6; // Set Data lines to HiZ (has to be after posedge)

//     INTERNAL SIGNALS
  reg  [`REG_WIDTH-1:0] StatusReg,  CntrReg;
  reg  [`REG_WIDTH-1:0] WriteOnlyCntrReg;
  reg  LocCS_Del_0, LocCS_Del_1, LocCS_Del_2; 
  
//  ACTUAL CODE STARTS HERE
  assign  ClockSelect        =  CntrReg[2];
  assign  RepeatMode         =  CntrReg[7];
  assign  PatternMode        =  CntrReg[8];
  assign  CounterModeSel     =  CntrReg[10];
  assign  GateTTCcommands    =  CntrReg[11];
  assign  StartStop_Spy      =  CntrReg[12];

  //  WriteOnlyCntrReg self-reset itself and creates positive pulses
  assign  VME_Hard_rst        =   WriteOnlyCntrReg[0];
  assign  VME_Soft_rst        =   WriteOnlyCntrReg[1];
  assign  VME_Start           =   WriteOnlyCntrReg[2];
  assign  VME_Stop            =   WriteOnlyCntrReg[3];
  assign  VME_AsyncFifoRst    =   WriteOnlyCntrReg[7]; 

// forcing IOBUFfer to convice XST.  LocData goes stright to the pads while the rest of
//  the logic makes use of Din and Dout. NB: T=1 means Tristate.
//    _S_2  means Slow output, 2 mA drive (minimum, in case of bus collisions)
 IOBUF_S_2 LocDataBUF0  (.IO(LocData[0]),  .I(Dout[0]),  .O(Din[0]),  .T(Tri));
 IOBUF_S_2 LocDataBUF1  (.IO(LocData[1]),  .I(Dout[1]),  .O(Din[1]),  .T(Tri));
 IOBUF_S_2 LocDataBUF2  (.IO(LocData[2]),  .I(Dout[2]),  .O(Din[2]),  .T(Tri));
 IOBUF_S_2 LocDataBUF3  (.IO(LocData[3]),  .I(Dout[3]),  .O(Din[3]),  .T(Tri));
 IOBUF_S_2 LocDataBUF4  (.IO(LocData[4]),  .I(Dout[4]),  .O(Din[4]),  .T(Tri));
 IOBUF_S_2 LocDataBUF5  (.IO(LocData[5]),  .I(Dout[5]),  .O(Din[5]),  .T(Tri));
 IOBUF_S_2 LocDataBUF6  (.IO(LocData[6]),  .I(Dout[6]),  .O(Din[6]),  .T(Tri));
 IOBUF_S_2 LocDataBUF7  (.IO(LocData[7]),  .I(Dout[7]),  .O(Din[7]),  .T(Tri));
 IOBUF_S_2 LocDataBUF8  (.IO(LocData[8]),  .I(Dout[8]),  .O(Din[8]),  .T(Tri));
 IOBUF_S_2 LocDataBUF9  (.IO(LocData[9]),  .I(Dout[9]),  .O(Din[9]),  .T(Tri));
 IOBUF_S_2 LocDataBUF10 (.IO(LocData[10]), .I(Dout[10]), .O(Din[10]), .T(Tri));
 IOBUF_S_2 LocDataBUF11 (.IO(LocData[11]), .I(Dout[11]), .O(Din[11]), .T(Tri));
 IOBUF_S_2 LocDataBUF12 (.IO(LocData[12]), .I(Dout[12]), .O(Din[12]), .T(Tri));
 IOBUF_S_2 LocDataBUF13 (.IO(LocData[13]), .I(Dout[13]), .O(Din[13]), .T(Tri));
 IOBUF_S_2 LocDataBUF14 (.IO(LocData[14]), .I(Dout[14]), .O(Din[14]), .T(Tri));
 IOBUF_S_2 LocDataBUF15 (.IO(LocData[15]), .I(Dout[15]), .O(Din[15]), .T(Tri));

 always@(posedge clk)  begin  
   if(rst==1)   begin       //   RESET
         SlbsID <=  'hCA;     
   end  //   END of RESET 
   else  begin   // posedge clk       
      LocCS_Del_0  <=  LocCS_b;    // FF to avoid metastability
      LocCS_Del_1  <=  LocCS_Del_0;
      LocCS_Del_2  <=  LocCS_Del_1;  // total of 3Tck = 75 ns delay on CS
      StatusReg[0]                <=  DLL_lock;
      StatusReg[1]                <=  run_status;  
      StatusReg[7:2]              <=  LastTTCcommand[7:2];
      StatusReg[8]                <=  0;
      StatusReg[9]                <=  0;
      StatusReg[12:10]            <=  UnlockCnt;
      StatusReg[13]               <=  BunchEr;
	   StatusReg[`REG_WIDTH-1:14]  <=  0;  // unused portion of status reg
      
    case(LB_STATE)
        
    IDLE:  // check for a negative edge on LocCS_b (digitally filtered)
        begin
          Dout      <=  'h BAD;  //mark a bad LocBus read timing
          Tri        <=  1;    // Data lines normally Hi-Z
          if(LocCS_Del_2 == 1 && LocCS_Del_1==0)  LB_STATE  <=  RD_OR_WR;
        end
            
    RD_OR_WR: //  decide whether it is a READ or a WRITE
        begin
          if (LocWRITE_b==0)
             LB_STATE <= MASTER_WRITE; //  Write operation (master's point of view)
          else  begin 
             LB_STATE <= MASTER_READ;  //  Read operation (master's point of view)
             Tri      <= 0;            //  Set Data lines toward LocalBus
            end  //  else
        end          
      //  ***********************************************   
      //     ********   W R I T E  operation  ********    
    MASTER_WRITE:  begin
       casez(LocAddr)    // decode which register is addressed                           
        'hC:     CntrReg           <=  Din[`REG_WIDTH-1:0];
        'h10:    WriteOnlyCntrReg  <=  Din[`REG_WIDTH-1:0];
        'h20:    SlbsID            <=  Din[`REG_WIDTH-1:0];     
        'h28:    ChMask[11:0]      <=  Din[`REG_WIDTH-1:0];
        'h2C:    ChMask[23:12]     <=  Din[`REG_WIDTH-1:0];
        'h30:    MaxBCN            <=  Din[`REG_WIDTH-1:0];                            
        default:       ;
               //no action  if there is no such register or is read-only 
       endcase    //    (LocAddr)
       LB_STATE  <=  END_OF_WRITE;          
    end        //  end of MASTER_WRITE case

          //  ****************************************
          //    ******* R E A D  operation ********
    MASTER_READ:      begin
          PatternsWrEn  <= 0;
       casez(LocAddr)                                    
              // FIFO section: generate Fifo_RD pulses,
              // the data will be available on the next cycle
/*        'hD0:   SnapRdEn[1]        <=  1'b1;      
        'hD4:   SnapRdEn[2]        <=  1'b1;      
        'hD8:   SnapRdEn[3]        <=  1'b1;      
        'hDC:   SnapRdEn[4]        <=  1'b1;      
        'hE0:   SnapRdEn[5]        <=  1'b1;      
        'hE4:   SnapRdEn[6]        <=  1'b1;      
        'hE8:   SnapRdEn[7]        <=  1'b1;      
        'hEC:   SnapRdEn[8]        <=  1'b1;      */
        default:       ;
       endcase   //  (LocAddr)
       LB_STATE  <=  END_OF_READ;          
      end      //  MASTER_READ  case
            
    END_OF_READ:  begin                   
       SnapRdEn[8:1]    <=  0;   

       casez(LocAddr)  // here we actually latch the data outputs 
        'h0:   Dout <=  'h0;      //  this reg is for testing
        'h4:   Dout <=  'hFFFFF;  //  this reg is for testing
        'h8:   Dout <=  {11'b0, LastTTCcommand};
        'hC:   Dout <=  CntrReg;
        'h14:  Dout <=  StatusReg;
        'h20:  Dout <=  SlbsID;
        'h24:  Dout <=  `FPGAversionN;
        'h28:  Dout <= {4'b0, ChMask[11:0]};
        'h2C:  Dout <= {4'b0, ChMask[23:12]};             
        'h30:  Dout <= {4'b0, MaxBCN};
        'h40:  Dout <= TTC_test0;
        'h44:  Dout <= TTC_test1;
        'h48:  Dout <= TTC_test2;
        'h4C:  Dout <= TTC_test3;
        'h50:  Dout <= TTC_test4;
        'h54:  Dout <= TTC_test5;
        'h58:  Dout <= TTC_test6;
        'h5C:  Dout <= TTC_test7;
        'h60:  Dout <= TTC_test8;
        'h64:  Dout <= TTC_test9;
        'h68:  Dout <= TTC_testA;
        'h6C:  Dout <= TTC_testB;
        'h70:  Dout <= TTC_testC;
        'h74:  Dout <= TTC_testD;
        'h78:  Dout <= TTC_testE;
        'h7C:  Dout <= TTC_testF;                           
        default:  Dout <=  'H DEAD;
            // if there is no such read-register answer: DEAD 
       endcase              
       if(LocCS_Del_1==1)  // level-sensitive to LocCS_b (may avoid lock)
            LB_STATE    <=  END_CYCLE;
    end   // END_OF_READ 
          
    END_OF_WRITE:
       begin    
          WriteOnlyCntrReg <=  0;   // self-reset to generate 1-tick pulses
          if(LocCS_Del_1==1)  // level-sensitive to LocCS_b (may avoid lock)
            LB_STATE    <=  END_CYCLE;
       end
          
    END_CYCLE:  
       begin         
             PatternsWrEn  <= 0;        LB_STATE      <= IDLE;
       end
    
    default: LB_STATE    <=  IDLE;

    endcase    //  end the case related to the LB_STATE variable
   end      // else posedge clk
   end        // always
endmodule